"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
import json
import sys
from datetime import datetime

import hvac
import requests
from dynaconf import LazySettings
from jinja2 import Environment
from jinja2 import FileSystemLoader
from loguru import logger

from .store_engine import StoreEngine
from .store_object import StoreObject


class VaultClient:
    """Client commands not available via hvac"""

    __vault_client: hvac.Client

    __headers = {'Content-Type': 'application/json',
                 'X-Vault-Token': ''}
    __url: str            #: Vault server URL
    __token: str          #: Auth token
    __approle: str        #: Application Role ID
    __secretid: str       #: Application Role Secret ID
    __mount: str          #: Engine mount path
    __mounts: json        #: Map of Vault mounts

    def __init__(self, settings: LazySettings) -> None:
        """Constructor for VaultClient"""
        self.__url = settings.VAULT_URL
        self.__approle = settings.VAULT_APPROLE
        self.__secretid = settings.VAULT_SECRET_ID
        self.__mount = settings.VAULT_MOUNT
        self.__mounts = {}
        self.__vault_client = hvac.Client(url=self.__url)
        self.__settings = settings

    def initialize(self) -> bool:
        if self.__settings.CTX.obj['ADMIN_MODE']:
            return self.new_mount(mount=self.mount)
        else:
            return self.connect()

    def connect(self) -> bool:
        try:
            logger.trace(f'Connecting to Vault approle: {self.__approle} secret_id: {self.__secretid}')
            resp = self.__vault_client.auth_approle(role_id=self.__approle, secret_id=self.__secretid, use_token=True)
            self.__token = resp['auth']['client_token']
            logger.trace(f'client_token: {self.__token}')
            self.__headers['X-Vault-Token'] = self.__token
        except requests.exceptions.ConnectionError as err:
            logger.error(f'Failed to connect to {self.__url}: {err}')
            sys.exit(2)
        except hvac.exceptions.VaultError as err:
            logger.error(f'Failed to authenticate with Vault {err}')
            sys.exit(2)
        else:
            return True

    def url(self) -> str:
        return self.__url

    @property
    def mount(self) -> str:
        return self.__mount

    @property
    def token(self) -> str:
        return self.__token

    def logout(self) -> bool:
        return self.__vault_client.logout()

    def _get(self, path: str) -> json:
        """GET REST API wrapper method

            :param path: Vault API to query
            :type path: String
            :return: JSON paylod
        """
        try:
            """Connect refreshes the temp Vault auth token"""
            self.connect()
            response = requests.get(self.__url+path, headers=self.__headers)
            response.raise_for_status()

            return json.loads(response.content.decode('utf-8'))

        except requests.exceptions.HTTPError as errh:
            logger.error(f'Http Error: {errh}')
        except requests.exceptions.ConnectionError as errc:
            logger.error(f'Error Connecting: {errc}')
        except requests.exceptions.Timeout as errt:
            logger.error(f'Timeout Error: {errt}')
        except requests.exceptions.RequestException as err:
            logger.error(f'Error: {err}')

    def _post(self, path: str, data: json) -> json:
        """POST REST API wrapper method

            :param path: Vault API to change or create
            :type path: String
            :param data: Required request body
            :type data: JSON
            :return: requests.Response object
        """
        try:
            """Connect refreshes the temp Vault auth token"""
            self.connect()
            response = requests.post(self.__url+path, headers=self.__headers, data=data)
            response.raise_for_status()

            return json.loads(response.content.decode('utf-8'))

        except requests.exceptions.HTTPError as errh:
            logger.error(f'Http Error: {errh}')
        except requests.exceptions.ConnectionError as errc:
            logger.error(f'Error Connecting: {errc}')
        except requests.exceptions.Timeout as errt:
            logger.error(f'Timeout Error: {errt}')
        except requests.exceptions.RequestException as err:
            logger.error(f'Error: {err}')

    def _put(self, path: str, data: json) -> requests.Response:
        """PUT REST API wrapper method

            :param path: Vault API to change or create
            :type path: String
            :param data: Required request body
            :type data: JSON
            :return: requests.Response object
        """
        try:
            """Connect refreshes the temp Vault auth token"""
            self.connect()
            response = requests.put(self.__url+path, headers=self.__headers, data=data)
            response.raise_for_status()

            return response

        except requests.exceptions.HTTPError as errh:
            logger.error(f'Http Error: {errh}')
        except requests.exceptions.ConnectionError as errc:
            logger.error(f'Error Connecting: {errc}')
        except requests.exceptions.Timeout as errt:
            logger.error(f'Timeout Error: {errt}')
        except requests.exceptions.RequestException as err:
            logger.error(f'Error: {err}')

    def get_mounts(self) -> json:
        """Refresh set of mounts from Vault

            :return: JSON
        """
        self.__mounts = self._get("/v1/sys/mounts")
        return self.__mounts

    def new_mount(self, mount: str) -> bool:
        """Will create a vault mount of type k/v V2 if it doesn't exist

            :param mount: Name of the Vault K/V Secret Engine
            :type mount: String
            :return: Boolean
        """
        self.__mounts = self._get("/v1/sys/mounts")
        logger.trace(f'{self.__class__}::new_mount {self.__mounts}')
        if self.__mounts and mount+"/" not in self.__mounts['data'].keys():
            logger.info(f'Vault mount {self.__url}/v1/sys/mounts/{mount} does not exist, creating')
            self._put(f'/v1/sys/mounts/{mount}', '{"type": "kv-v2"}')
            self.__mounts = self._get("/v1/sys/mounts")
        return bool(self.__mounts)

    def upsert(self, obj: StoreObject) -> bool:
        client = self.__vault_client
        mp = self.mount
        policyname = ""

        try:
            self.connect()
            logger.trace(f'updating secret {mp}/{obj.path_name}/cert_body')
            client.secrets.kv.v2.create_or_update_secret(path=obj.path_name + "/cert_body",
                                                         mount_point=mp,
                                                         secret=obj.data['cert_body'])

            self.connect()
            logger.trace(f'updating secret {mp}/{obj.path_name}/cert_info')
            client.secrets.kv.v2.create_or_update_secret(path=obj.path_name + "/cert_info",
                                                         mount_point=mp,
                                                         secret=obj.data['cert_info'])
            self.connect()
            list_policies_resp = client.sys.list_policies()['data']['policies']
            commonname = obj.data['cert_info']['subject']['commonName']
            policyname = f'knox-read-{commonname}'
            if policyname in list_policies_resp:
                pass
            else:
                policy = obj.data['cert_policy']
                logger.debug(f'Creating explict read access policy {policyname} for {mp}{obj.path_name}/cert_body')
                logger.trace(f'{self.__class__}::upsert {policyname}:\n{policy}')
                self.connect()
                client.sys.create_or_update_policy(name=policyname, policy=policy)

        except hvac.exceptions.Forbidden as ve:
            policyphrase = ""
            if len(policyname) > 0:
                policyphrase = f'and or {policyname}'
            logger.error(f'Permission denied writing {obj.path_name} {policyphrase}: {ve}')
            sys.exit(2)
        except hvac.exceptions.InvalidPath as ve:
            policyphrase = ""
            if len(policyname) > 0:
                policyphrase = f'and or {policyname}'
            logger.error(f'Path invalid for {obj.path_name} {policyphrase}: {ve}')
            sys.exit(2)
        except hvac.exceptions.Unauthorized as ve:
            policyphrase = ""
            if len(policyname) > 0:
                policyphrase = f'and or {policyname}'
            logger.error(f'Credentials not authorized to write {obj.path_name} {policyphrase}: {ve}')
            sys.exit(2)
        except Exception as vex:
            logger.error(f'Failed to write StoreObject to Vault {vex}')
            sys.exit(2)
        else:
            logger.info(f'Successfully stored {obj.path_name} and {policyname}')
            return True

    def read(self, path: str, name: str, type=None) -> tuple:
        client = self.__vault_client
        if type:
            fullpathbody = f'{path}/{name}/{type}/cert_body'
            fullpathinfo = f'{path}/{name}/{type}/cert_info'
        else:
            fullpathbody = f'{path}/{name}/cert_body'
            fullpathinfo = f'{path}/{name}/cert_info'

        try:
            self.connect()
            logger.trace(f'Attempting to read \n\tbody:{fullpathbody}\n\tinfo:{fullpathinfo}')
            logger.trace(f'client.url: {client.url}')
            logger.trace(f'mount: {self.mount}')
            certbody = client.secrets.kv.v2.read_secret_version(path=fullpathbody, mount_point=self.mount)
            self.connect()
            certinfo = client.secrets.kv.v2.read_secret_version(path=fullpathinfo, mount_point=self.mount)
            return certbody, certinfo

        except hvac.exceptions.Forbidden as ve:
            logger.error(f'Permission denied reading {path}/{name}: {ve}')
            sys.exit(2)
        except hvac.exceptions.InvalidPath as ve:
            logger.error(f'Path invalid for {path}/{name}: {ve}')
            sys.exit(2)
        except hvac.exceptions.Unauthorized as ve:
            logger.error(f'Credentials not authorized to read {path}/{name}: {ve}')
            sys.exit(2)

    def search(self, rootpath: str, rootkey: str, searchresults: list) -> list:
        """Search for 'cert_info' for a given vault path

            :param rootpath: Beginning search path
            :type rootpath: str
            :param rootkey: Used to get commonname from search path
            :type str:
            :param searchresults: Stores the search results..default is empty
            :type list:
            :return: list
        """
        try:
            client = self.__vault_client
            self.connect()
            rootpath = rootpath.replace('//', '/')
            path = rootpath
            logger.trace(f'Searching {path}')
            secrets = client.secrets.kv.list_secrets(path=path, mount_point=self.mount)
            secrets_keys = secrets.get('data').get('keys')
            if isinstance(secrets_keys, list):
                if 'cert_info' not in secrets_keys:
                    for key in secrets_keys:
                        subpaths = rootpath + key
                        self.search(rootpath=subpaths, rootkey=key, searchresults=searchresults)
                else:
                    cert_info_path = rootpath + "cert_info"
                    cert_common_name = rootpath.split('/')[-3]
                    self.connect()
                    cert_info_dict = client.secrets.kv.v2.read_secret_version(path=cert_info_path,
                                                                              mount_point=self.mount)
                    current_date = datetime.now()
                    cert_expiry_date = datetime.strptime(cert_info_dict.get('data').get('data')
                                                         .get('validity').get('not_valid_after'), '%Y-%m-%d %H:%M:%S')
                    days_to_expire = cert_expiry_date - current_date
                    results_dict = {'common_name': cert_common_name, 'vault_cert_path': rootpath,
                                    'cert_issue_date': (cert_info_dict.get('data').get('data').get('validity')
                                                        .get('not_valid_before')),
                                    'cert_expiry_date': (cert_info_dict.get('data').get('data').get('validity')
                                                         .get('not_valid_after')),
                                    'days_to_expire': days_to_expire.days}
                    searchresults.append(results_dict)
        except requests.exceptions.ConnectionError as ve:
            logger.error(f'Failed to connect to {self.__url}: {ve}')
            sys.exit(2)
        except hvac.exceptions.Forbidden as ve:
            logger.error(f'Permission denied for reading from {rootpath}: {ve}')
            sys.exit(2)
        except hvac.exceptions.InvalidPath as ve:
            logger.error(f'Path not found for {rootpath}: {ve}')
            sys.exit(2)
        except hvac.exceptions.Unauthorized as ve:
            logger.error(f'Credentials not authorized to access {rootpath}: {ve}')
            sys.exit(2)
        else:
            return searchresults


class VaultStoreEngine(StoreEngine):
    """Vault implementation of the StoreEngine interface"""
    __client: VaultClient

    def __init__(self, settings) -> None:
        """Constructor for VaultStoreEngine"""
        super().__init__()
        self._settings = settings
        self.__client = VaultClient(settings)
        logger.debug(f'🔐 Vault backend configuration loaded. {self.__client.url()}')
        if self.initialize():
            logger.debug("🔐 Vault backend initialized.")
        else:
            logger.error(f'🛑 Failed to connect to Vault {self.__client.url()}')

    def initialize(self) -> bool:
        return self.__client.initialize()

    def open(self) -> bool:
        """Ensure the vault client is connected"""
        return self.__client.connect()

    def close(self) -> bool:
        """Ensure we close the vault connection"""
        return self.__client.logout()

    def __del__(self):
        """When the object is destroyed make sure we close the connection to Vault"""
        return self.close()

    def write(self, obj: StoreObject) -> bool:
        """Given a StoreObject, store it into vault using mount/path/name == body,info

            :param obj: The StoreObject to persist
            :type obj: StoreObject
            :return: bool
        """
        return self.__client.upsert(obj)

    def read(self, path: str, name: str, type=None) -> StoreObject:
        """Using the provided path and name retrieve the data from the store and create a new StoreObject

            :param path: Store path to the object
            :type path: str
            :param name: Name of the object to retrieve
            :type name: str
            :param type: StoreObject type, if known
            :type type: str
            :return: StoreObject
        """
        try:
            certbody, certinfo = self.__client.read(path, name, type)
            cert = StoreObject(name=name,
                               path=path,
                               body=certbody['data']['data'],
                               info=certinfo['data']['data'],
                               type=type)
            cert._data = {'cert_body': certbody['data']['data'],
                          'cert_info': certinfo['data']['data']}

        except Exception as vex:
            logger.error(f'Failed to read StoreObject /{self.__client.mount()}{path}/{name} {vex}')
            sys.exit(2)
        else:
            logger.info(f' Successfully read {cert.path_name}')
            return cert

    def find(self, pattern) -> list:
        """Search certificate info for a given search pattern

            :param pattern: Search glob pattern
                ex: abc.8x8.com, abc.8x8.com/*, 8x8.com/*
            :type pattern: str

            :return: list
        """
        searchpath = '/'.join(reversed(pattern.strip('/*').split('.'))) + "/"
        return self.__client.search(rootpath=searchpath, rootkey="", searchresults=[])



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import ast
import sys

import boto3
from botocore.exceptions import ClientError
from botocore.exceptions import EndpointConnectionError
from dynaconf import LazySettings
from jinja2 import Environment
from jinja2 import FileSystemLoader
from loguru import logger

from .store_engine import StoreEngine
from .store_object import StoreObject


class ACMStoreEngine(StoreEngine):
    """
    ACMStoreEngine Class
    CRUD operations for Aws Certificate Manager
    """

    __AwsErrors = (ClientError, EndpointConnectionError)
    __session: boto3.session.Session

    def __init__(self, settings: LazySettings = None):
        super().__init__()
        self.profile_name = settings.AWS_PROFILE
        self.region = settings.AWS_REGION
        self.CertArn = None
        self.__path = None
        self._jinja = Environment(loader=FileSystemLoader('templates'))
        self._tmpl_tags = self._jinja.get_template('tags_template.js')
        if self.initialize():
            logger.debug('Connected to ACM')
        else:
            logger.error('No AWS profiles found')
            sys.exit(1)

    def initialize(self) -> bool:
        self.__session = boto3.Session(profile_name=self.profile_name, region_name=self.region)
        return len(self.__session.available_profiles) > 0

    def find(self, pattern: str) -> [StoreObject]:
        certs = []
        try:
            acm_res = self.__session.client('acm').list_certificates(
                CertificateStatuses=['ISSUED'],
                MaxItems=123
            )
            acm_certs = acm_res.get('CertificateSummaryList')
            for acm in acm_certs:
                domainname = acm['DomainName']
                arn = acm['CertificateArn']
                if domainname == pattern:
                    certs.append(self.get(name=domainname, arn=arn))

        except self.__AwsErrors as e:
            logger.error(f'[AWSCert]: Exception listing certificates from ACM {e}')
            sys.exit(1)
        else:
            return certs

    def get(self, name: str, arn: str) -> StoreObject:
        try:
            acm_res = self.__session.client('acm').get_certificate(
                CertificateArn=arn
            )
            cert = StoreObject(name=name)
            cert.public = acm_res.get('Certificate')
            if hasattr(acm_res, 'CertificateChain'):
                cert.chain = acm_res.get('CertificateChain')

        except self.__AwsErrors as e:
            logger.error(f'[AWSCert]: Exception listing certificates from ACM {e}')
            sys.exit(1)
        return cert

    def read(self, name: str, path: str = None, type=None) -> StoreObject:
        certs = self.find(name)
        for cert in certs:
            if cert.name == name:
                return cert

    def write(self, cert: StoreObject) -> bool:
        """ ACM Store Engine Write the certificate to specified region and account

            :param cert: The StoreObject to persist in AWS ACM Store
            :type cert: Cert
            :return: bool

        """
        logger.trace(f'[ACMStoreEngine]:\nPUB  :{cert.public}\nKEY  :REDACTED\nCHAIN:{cert.chain}\n')

        try:
            if len(cert.data['cert_body']['chain']) > 0:
                acm_res = self.__session.client('acm').import_certificate(
                    Certificate=cert.data['cert_body']['public'],
                    PrivateKey=cert.data['cert_body']['private'],
                    CertificateChain=cert.data['cert_body']['chain'],
                    Tags=ast.literal_eval(self._tmpl_tags.render(cert=cert))
                )
            else:
                acm_res = self.__session.client('acm').import_certificate(
                    Certificate=cert.data['cert_body']['public'],
                    PrivateKey=cert.data['cert_body']['private'],
                    Tags=ast.literal_eval(self._tmpl_tags.render(cert=cert))
                )

            cert.arn = acm_res.get('CertificateArn')
            logger.trace(
                f'[ACMStoreEngine]: Certificate uploaded:\n'
                f'Region: {self.region}\n'
                f'Account: {self.profile_name}\n'
                f'CertARN: {cert.arn}')
            return True

        except self.__AwsErrors as e:
            logger.error(f'[ACMStoreEngine]: Exception importing certificates to ACM {e}')
            sys.exit(1)



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
from loguru import logger

from .store_acm import ACMStoreEngine
from .store_engine import StoreEngine
from .store_file import FileStoreEngine
from .store_object import StoreObject
from .store_vault import VaultStoreEngine


class Store:
    """Abstract class to generalize access to the different stores"""
    _engine: StoreEngine
    _engine_name: str
    _engine_map = {
        'vault': VaultStoreEngine,
        'file': FileStoreEngine,
        'aws': ACMStoreEngine
    }

    def __init__(self, settings, engine_name: str = None) -> None:
        """Dynamically load StoreEngine type from .env via Dynaconf
        KNOX_STORE_ENGINE=[vault,file,aws]

        :param settings: Dynaconf LazySettings
        :type settings: dynaconf.LazySettings
        :param engine_name: String indicating which StoreEngine to instantiate
        :type engine_name: String
        """
        try:
            self._engine_name = engine_name if engine_name is not None else settings.STORE_ENGINE
            self._engine = self._engine_map.get(self._engine_name).__call__(settings)
        except Exception:
            logger.error(
                f'StoreEngineFailure KNOX_STORE_ENGINE={self._engine_name} is invalid. Valid options are {self._engine_map.keys()}')  # noqa: E501
            raise

        self._engine.settings = settings
        logger.debug(f'Loaded {self._engine.__class__}')

    def save(self, obj: StoreObject) -> bool:
        """Save the given object to persistence"""
        return self._engine.write(obj)

    def get(self, path: str, name: str, type=None) -> StoreObject:
        """Given path read object"""
        return self._engine.read(path, name, type)

    def delete(self, path: str, name: str) -> bool:
        """Remove the object from the store"""
        """[TODO 5/13/20] ljohnson implement soft delete and hard deletes"""
        return self._engine.delete(path, name)

    def find(self, pattern: str) -> list:
        """Given a pattern, return collection of all objects
           Search patterns : abc.8x8.com, abc.8x8.com/*, 8x8.com/*
        """
        return self._engine.find(pattern)



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. """
import csv
import json
import sys

import click
import pkg_resources
from loguru import logger

from .certificate import Cert  # noqa: F401
from .config import Conf
from .knox import Knox


@click.group()
@click.option("--log", "-l",
              type=click.Choice(['TRACE',
                                 'DEBUG',
                                 'INFO',
                                 'SUCCESS',
                                 'WARNING',
                                 'ERROR',
                                 'CRITICAL']),
              default='INFO',
              show_default=True,
              help="Sets the level of logging displayed")
@click.option("--verbose", "-v", is_flag=True, default=False, help="Display log output to console")
@click.option("--admin", is_flag=True, default=False, help="Enable admin actions", hidden=True)
@click.version_option(version=pkg_resources.get_distribution('knox').version)
@click.pass_context
@logger.catch()
def cli(ctx, verbose: bool = False, log: str = 'INFO', admin: bool = False):
    """Utilities for managing and storing TLS certificates using backing store (Vault)."""
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['LOG_LEVEL'] = log
    ctx.obj['ADMIN_MODE'] = admin
    logger.remove()
    if verbose:
        logger.add(sys.stdout,
                   format="{time} {level: >9} {level.icon} {message}",
                   filter=Conf.log_filter,
                   level=f"{log}",
                   colorize=True)
        logger.info(f' Log level set to {ctx.obj["LOG_LEVEL"]}')
    else:
        logger.add(sys.stderr,
                   format="{time} {level: >9} {level.icon} {message}",
                   filter=Conf.log_filter,
                   level=f"{log}",
                   colorize=True)


@cli.group(no_args_is_help=True)
@click.option("--type", "-t",
              type=click.Choice(['PEM', 'DER', 'PFX'], case_sensitive=False),
              default='PEM',
              show_default=True)
@click.option("--pub", help="Public key file")
@click.option("--chain", help="Intermediate chain")
@click.option("--key", help="Private key file")
@click.pass_context
@logger.catch()
def cert(ctx, pub: str, key: str, type: str = 'PEM', chain: str = None):
    """Certificate utilities.

    NAME is the common name for the certificate. i.e. www.example.com
    """
    ctx.obj['CERT_PUB'] = pub
    ctx.obj['CERT_CHAIN'] = chain
    ctx.obj['CERT_KEY'] = key
    ctx.obj['CERT_TYPE'] = type.upper()


@cert.command(name="save", no_args_is_help=True)
@click.argument("name")
@click.pass_context
@logger.catch()
def cert_save(ctx, name):
    """Store an existing certificate
    """
    ctx.obj['CERT_NAME'] = name
    pub = ctx.obj['CERT_PUB']
    key = ctx.obj['CERT_KEY']
    chain = ctx.obj['CERT_CHAIN']
    certtype = ctx.obj['CERT_TYPE']

    knox = Knox(ctx)
    certificate = Cert(knox.settings, common_name=name)
    certificate.load(pub=pub,
                     key=key,
                     chain=chain,
                     certtype=certtype)
    knox.store.save(certificate)


@cert.command(name="get", no_args_is_help=True)
@click.argument("name")
@click.pass_context
@logger.catch()
def cert_get(ctx, name):
    """Retrieve an existing certificate for a given common name
    """
    ctx.obj['CERT_NAME'] = name
    knox = Knox(ctx)
    certificate = Cert(knox.settings, common_name=name)
    certificate.type = ctx.obj['CERT_TYPE']
    certificate = knox.store.get(certificate.store_path(), name=name, type=certificate.type)
    with open(certificate.name+"-pub.pem", "w") as pubf:
        pubf.write(certificate.body['public'])
    with open(certificate.name+"-key.pem", "w") as keyf:
        keyf.write(certificate.body['private'])
    with open(certificate.name+"-chain.pem", "w") as chainf:
        chainf.write(certificate.body['chain'])


@cert.command(name="gen", no_args_is_help=True)
@click.argument("name")
@click.pass_context
@logger.catch()
def cert_gen(ctx, name):
    """Create and store a new certificate for a given common name
    """
    ctx.obj['CERT_NAME'] = name

    knox = Knox(ctx)
    certificate = Cert(knox.settings, common_name=name)
    certificate.generate()
    knox.store.save(certificate)


@cert.command(name="aws", no_args_is_help=True)
@click.argument("name")
@click.option("--region", default=None, help="Default AWS region")
@click.option("--profile", default=None, help="Default AWS profile to use")
@click.pass_context
@logger.catch()
def cert_aws(ctx, name, region, profile):
    """Store a certificate for a given common name in AWS
    """

    ctx.obj['CERT_NAME'] = name
    pub = ctx.obj['CERT_PUB']
    key = ctx.obj['CERT_KEY']
    chain = ctx.obj['CERT_CHAIN']
    certtype = ctx.obj['CERT_TYPE']

    knox = Knox(ctx)
    certificate = Cert(knox.settings, common_name=name)
    certificate.load(pub=pub, key=key, chain=chain, certtype=certtype)
    knox.attach("aws")
    # Save to the attached AWS ACM store
    knox.stores("aws").save(certificate)
    # Save to the default store
    knox.store.save(certificate)


@cli.group(no_args_is_help=True)
@click.pass_context
def store(ctx) -> dict:
    """Store commands. Given a certificate NAME perform the store operation."""
    pass


@store.command(name="find", no_args_is_help=True)
@click.option("--file", "-f", help="Output file, default stdout")
@click.option("--output", "-o",
              type=click.Choice(['JSON', 'CSV'], case_sensitive=False),
              default='JSON',
              show_default=True,
              help="Type of output")
@click.argument("name")
@click.pass_context
@logger.catch()
def store_find(ctx, name, file: str = 'stdout', output: str = 'JSON') -> dict:
    """Given a certificate NAME pattern search the store.

    NAME can be similar to a full file path or the certificates common name.
    i.e. www.example.com becomes /com/example/www/www.example.com when stored.
    Supports wild cards. *.example.com

    """
    ctx.obj['STORE_FIND_NAME'] = name
    ctx.obj['STORE_FIND_OUTPUT'] = output
    ctx.obj['STORE_FIND_OUTFILE'] = file
    knox = Knox(ctx)
    if name:
        results = knox.store.find(pattern=name)  # noqa F841
        handle = open(file, 'w') if file else sys.stdout
        if output == 'JSON':
            handle.write(json.dumps(results, indent=4))
        else:
            csv_writer = csv.writer(handle)
            count = 0
            for rec in results:
                if count == 0:
                    header = rec.keys()
                    csv_writer.writerow(header)
                    count += 1
                csv_writer.writerow(rec.values())

        handle.close()

    return ctx


def main():
    cli(prog_name="knox", obj={})


if __name__ == "__main__":
    main()
